// ********************* USB / PS2 Keyboard Decoder ****************************
//
// With this program we decode 1 or 2 Keyscan and translate them
// to useable ASCII code
// After translating the code is send to uart
//
// Microcontroller    : PIC16F688
//                    : 8Mhz Internal Oscillator
//
// We use software uart that we can invert the signals if needed (max232)
//
// Connect C4 as TXD            to RS232 RXD Connector
// Connect C5 as RXD (not used) to RS232 TXD Connector
// Connect the keyboard clock output to pin C0
// Connect the keyboard data  output to pin C1
//
// A jumper will set the Keyboard country
// The jumper have to be connected betwen +5v and pin A0 or pin A1
// No Jumper = German  Keyboard
// A0        = French  Keyboard
// A1        = English Keyboard
//
// Programmer              : F. Muller
// Compiler                : CCS C v4.114
// Version                 : 1.4
// Date                    : 15 Apr 2011
// *****************************************************************************

#include <16F688.h>
#include "keyboard.h"

//~~~ 8MHZ OSCILLATOR CONFIGS ~~~//
#FUSES NOWDT,INTRC_IO,NOPROTECT,NOBROWNOUT,MCLR,NOCPD,PUT,NOIESO,FCMEN
#use delay(clock=8000000)
// Here there a 2 rs232 communication possibility
// Both are do in software
// The first is set as normal rs232 stream
// The secont is set as inverted signal for some case like Tinydisplay or TinyBrick
#use rs232(baud=9600, xmit=PIN_C4, rcv=PIN_C5, FORCE_SW, BITS=8, STOP=1, PARITY=N)
//#use rs232(baud=9600, xmit=PIN_C4, rcv=PIN_C5, FORCE_SW, BITS=8, STOP=1, PARITY=N, INVERT)
// rs232 for Direct debug to PIC Kit 2
//#use rs232(baud=9600, xmit=PIN_A0, rcv=PIN_A1, FORCE_SW, BITS=8, STOP=1, PARITY=N)

// Keyboard Clock
#ifndef PIN_CLK
#define PIN_CLK PIN_C0
#endif

// Keyboard Data
#ifndef PIN_DATA
#define PIN_DATA PIN_C1
#endif

unsigned int8 ClockPulse1,ClockPulse2;    //Clock Pulses for keyscan 1 and 2
unsigned int8 DataInput1,DataInput2;      //Data get from keyscan code 1 and 2
unsigned int8 Shifted_Key;                //Shift key was press
unsigned int8 Alt_Gr_Key;                 //Alt-Gr key was press
unsigned int8 exitfunction;               //temp variable
unsigned int8 Translated;                 // = 1 if we found the character
char ToRS232;                             //The character to output to rs232
unsigned int8 KB_Country;                 //0 = German , 1 = French , 2 = English
unsigned int8 DebugMode;                  //To transfer all usefull data to rs232
unsigned int8 Counter;

//Functions
void PulseDecoder(void);
void Translate_Key(void);
void GetNormal_Key(void);
void GetShifted_Key(void);
void GetAltGr_Key(void);

void main()
{

   setup_adc_ports(NO_ANALOGS|VSS_VDD);
   setup_adc(ADC_OFF);
   setup_timer_0(RTCC_INTERNAL|RTCC_DIV_1);
   setup_comparator(NC_NC_NC_NC);
   setup_vref(FALSE);
   setup_oscillator(OSC_8MHZ);
   setup_timer_1(T1_INTERNAL);       //8Mhz 2000 ticks = 1Ms
                                     //        1 ticks = 500nS
   set_tris_c(227);                  //PortC setup
   set_tris_a(255);                  //PortA as input for Keyboard Country jumper
   
   // Delay to let the Keyboard initialize
   delay_ms(5000);      // Delay for keyboard init (5 seconds)
   
   // init some variables
   ClockPulse1 = 0;
   ClockPulse2 = 0;
   Shifted_Key = 0;
   Alt_Gr_Key = 0;
   ToRS232 = 0;
   
   // Keyboard Country
   KB_Country =0;                               // German Keyboard as standard
   DebugMode = 0;
   
   while(1)
   {
     //Wait for Keyboard Clock and scan the first Scancode
     while(input(PIN_CLK)) continue;                    // wait clock = 0
     //Reset variables at each cycle
     ClockPulse1 = 0;
     ClockPulse2 = 0;
     DataInput1 = 0;
     DataInput2 = 0;
     ToRS232 = 0;
     // we go decode all pulse comming from keyboard
     PulseDecoder();
     // Now we send out the 2 Keyscan if anything was read
     if (ClockPulse1 == 11)
      {
         Translated = 0;
         exitfunction = 0;
         // here we translate the keyscan to ascii
         Translate_Key();
         
         // if the key is know we output the result to rs232
         if (DebugMode == 1)
          {
            // Data Debug Output
            printf("key = %c : d1 = %u : d2 = %u : Shift = %u : AltGr = %u\n",ToRS232,DataInput1,DataInput2,Shifted_Key,Alt_Gr_Key);
            //printf("d1 = %u : d2 = %u\n",DataInput1,DataInput2); 
          }
         if ((Translated == 1) && (DebugMode == 0))
          {
            // Out to PC or TinyBasic CPU
            printf("%c",ToRS232);
          }
         delay_ms(10);  // Debounce delay
      }
   }
}

void PulseDecoder(void)
// Here we decode every keyscan
{
      set_timer1(0);                                     // reset timer1
      while ((ClockPulse1 < 11) && (get_timer1() < 2000))
      {
         if ((ClockPulse1 > 0) && (ClockPulse1 < 9))
         // Record the actual keyscan bit
         {
            if(input(PIN_DATA))
               {bit_set(DataInput1,ClockPulse1-1);}
            else
               {bit_clear(DataInput1,ClockPulse1-1);}
         }
         //wait for clock going to 1
         while (!input(PIN_CLK) && (get_timer1() < 2000)) continue;// wait clock = 1
         while ( input(PIN_CLK) && (get_timer1() < 2000)) continue;// wait clock = 0
         //increment clock counter
         ClockPulse1++;
      }
      // Delay to read the second keyscan code if exist
      Delay_ms(1);
      
      //Wait for Keyboard Clock and scan the second Scancode
      //We made a security to 5Ms
      //The maximum wait time betwen a double scancode is 2.8mS
      while(input(PIN_CLK) && (get_timer1() < 10000)) continue;    // wait clock = 0
      // Scan the second code
      // Note that this part will never run if the timer1 is > 10000 (5Ms)
      while ((ClockPulse2 < 11) && (get_timer1() < 8000))
      {
         if ((ClockPulse2 > 0) && (ClockPulse2 < 9))
         // Record the actual keyscan bit
         {
            if(input(PIN_DATA))
               {bit_set(DataInput2,ClockPulse2-1);}
            else
               {bit_clear(DataInput2,ClockPulse2-1);}
         }
         //wait for clock going to 1
         while (!input(PIN_CLK) && (get_timer1() < 8000)) continue;
         while ( input(PIN_CLK) && (get_timer1() < 8000)) continue;
         //increment clock counter
         ClockPulse2++;
      }
}

void Translate_Key(void)
{
   // First we test if the Shift or Alt-Gr key was pressed
   // then we have to look if these keys was released
   
   // Check if release the Shift Key
   if (Shifted_Key == 1)
    {
      if ((DataInput1 == Left_Shift[2]) && (DataInput2 == Left_Shift[3]))
         {
           Shifted_Key = 0;               //L-Shift released
           exitfunction = 1;
         }
      if ((DataInput1 == Right_Shift[2]) &&(DataInput2 == Right_Shift[3]))
         {
           Shifted_Key = 0;               //R-Shift released
           exitfunction = 1;
         }
    }
   // Check if release the Alt-Gr Key
   if (Alt_Gr_Key == 1)
    {
      if ((DataInput1 == Alt_Gr[2]) && (DataInput2 == Alt_Gr[3]))
         {
           Alt_Gr_Key = 0;                //  Alt-Gr released
           exitfunction = 1;
         }
    }

   // Now we can look if the L-Shift , R-Shift or Alt-Gr key are pressed
   if ((Shifted_Key == 0) && (exitfunction == 0))
    {
      if ((DataInput1 == Left_Shift[0]) && (DataInput2 == Left_Shift[1]))
         {
           Shifted_Key = 1;               //L-Shift pressed
           exitfunction = 1;
         }
      if ((DataInput1 == Right_Shift[0]) &&(DataInput2 == Right_Shift[1]))
         {
           Shifted_Key = 1;               //R-Shift pressed
           exitfunction = 1;
         }
    }
   if ((Alt_Gr_Key == 0) && (exitfunction == 0))
    {
      if ((DataInput1 == Alt_Gr[0]) && (DataInput2 == Alt_Gr[1]))
         {
           Alt_Gr_Key = 1;                //  Alt-Gr pressed
           exitfunction = 1;
         }
    }
    
   // Here we start translating the Keycode
   
   // Normal keys
   if ((Shifted_Key == 0) && (Alt_Gr_Key == 0) && (exitfunction == 0))
    {
     Counter = 0;
     GetNormal_Key();
    }
   
   // Shifted keys
   if ((Shifted_Key == 1) && (Alt_Gr_Key == 0) && (exitfunction == 0))
    {
     Counter = 0;
     GetShifted_Key();
    }

   // Alt-Gr keys
   if ((Shifted_Key == 0) && (Alt_Gr_Key == 1) && (exitfunction == 0))
    {
     Counter = 0;
     GetAltGr_Key();
    }
}

void GetNormal_Key(void)
// Translate Keycode without Shift or Alg-Gr into ASCII code
{
     while ((Counter <= NbNormalKey) && (exitfunction == 0))
      {
       switch (KB_Country)
        {
         case 0:  // German Keyboard
          {
            if ((DataInput1 == NormalKey[Counter][0]) && (DataInput2 == NormalKey[Counter][1]))
               {
                  ToRS232 = NormalKey[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
         case 1:  // French Keyboard
          {
            if ((DataInput1 == NormalKeyFR[Counter][0]) && (DataInput2 == NormalKeyFR[Counter][1]))
               {
                  ToRS232 = NormalKeyFR[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
         case 2:  // English Keyboard
          {
            if ((DataInput1 == NormalKeyEN[Counter][0]) && (DataInput2 == NormalKeyEN[Counter][1]))
               {
                  ToRS232 = NormalKeyEN[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
        }
         Counter++;
      }
}

void GetShifted_Key(void)
// Translate Keycode with Shift pressed into ASCII code
{
     while ((Counter <= NbShiftedKey) && (exitfunction == 0))
      {
       switch (KB_Country)
        {
         case 0:  // German Keyboard
          {
            if ((DataInput1 == ShiftedKey[Counter][0]) && (DataInput2 == ShiftedKey[Counter][1]))
               {
                  ToRS232 = ShiftedKey[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
         case 1:  // French Keyboard
          {
            if ((DataInput1 == ShiftedKeyFR[Counter][0]) && (DataInput2 == ShiftedKeyFR[Counter][1]))
               {
                  ToRS232 = ShiftedKeyFR[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
         case 2:  // English Keyboard
          {
            if ((DataInput1 == ShiftedKeyEN[Counter][0]) && (DataInput2 == ShiftedKeyEN[Counter][1]))
               {
                  ToRS232 = ShiftedKeyEN[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
        }
         Counter++;
      }
}

void GetAltGr_Key(void)
// Translate Keycode with Alg-Gr pressed into ASCII code
{
     while ((Counter <= NbAltGrKey) && (exitfunction == 0))
      {
       switch (KB_Country)
        {
         case 0:  // German Keyboard
          {
            if ((DataInput1 == AltGrKey[Counter][0]) && (DataInput2 == AltGrKey[Counter][1]))
               {
                  ToRS232 = AltGrKey[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
         case 1:  // French Keyboard
          {
            if ((DataInput1 == AltGrKeyFR[Counter][0]) && (DataInput2 == AltGrKeyFR[Counter][1]))
               {
                  ToRS232 = AltGrKeyFR[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
         case 2:  // English Keyboard
          {
            if ((DataInput1 == AltGrKeyEN[Counter][0]) && (DataInput2 == AltGrKeyEN[Counter][1]))
               {
                  ToRS232 = AltGrKeyEN[Counter][2];
                  Translated = 1;
                  exitfunction = 1;
               }
            break;
          }
        }
         Counter++;
      }
}


